import pickle
import re
from nltk import ngrams


def read(path):
    data = []
    with open(path, "r") as f:
        for line in f:
            data.append(line.replace('\n',''))
    return data

def save(data,path):
    textfile = open(path, "w")
    for element in data:
        textfile.write(element + "\n")


def read_dict(path):
    with open(path,'rb') as file:
        data = pickle.load(file)
    return data
    
def save_dict(data,path):
    with open(path, 'wb') as file:
        pickle.dump(data, file)

def get_ngram(sentence, n = 1):
    # sentence = tokenize(sentence)
    return ngrams(sentence.split(' '), n)

def tokenize(sentence):
    sent = sentence.replace('\n','').strip()
    sent = re.sub('(?<! )(?=[.,!?()])|(?<=[.,!?()])(?! )', r' ', sent)
    return sent


def padding_punct(s):
    s = s.replace('\n', '').strip()
    s = re.sub('([.,!?()]),"")', r' \1 ', s)
    s = re.sub('\s{2,}', ' ', s)
    return s




import datetime
import itertools
import pandas as pd

from termcolor import colored
from nltk import text
from tqdm import tqdm
from vi_nlp_core.ner.extractor import Extractor
from vi_nlp_core.testcase.tc_time import get_time_tc
from vi_nlp_core.testcase.tc_pername import get_person_name_tc
from vi_nlp_core.testcase.tc_date import get_date_test_case
from vi_nlp_core.utils.util import read
from vi_nlp_core.utils.preproces import Preprocess

VLSP_PATH = './vi_nlp_core/testcase/vlsp2016'
class Tester:
    def __init__(self):
        self.extractor = Extractor(load_dict=False)
        self.per_tc, self.per_expected_result = get_person_name_tc()
        self.time_tc, self.time_expected_result, self.time_spec_tc, self.time_spec_expected_result = get_time_tc()
        self.date_tc, self.date_expected_result = get_date_test_case()
        self.preprocessor = Preprocess()
        self.error_factor = 2

    def test_person_name(self):
        tc_num = 1
        total_acc = 0
        total_len = 0
        for tc, ex_result in zip(self.per_tc, self.per_expected_result):
            acc = 0
            for test, label in zip(tc, ex_result):
                pred = self.extractor.extract_person_name(
                    test, mode='pattern', rt='relative')['entities']
                if pred[0]['value'] in label:
                    acc += 1
                # print(pred[0]['value'], label , acc)
            self.write_result(tc_num, acc, len(tc))
            tc_num += 1
            total_len += len(tc)
            total_acc += (acc)
        self.write_result('Total', total_acc, total_len)

    def test_person_name_vlsp(self,mode='train'):
        vlsp = pd.read_csv(f"{VLSP_PATH}/{mode}_processed.csv")
        per_tc = vlsp['src']
        per_expected_result = vlsp['trg']

        correct = 0

        for test, label in zip(per_tc, per_expected_result):
            pred = self.extractor.extract_person_name(
                test, mode='pattern', rt='relative')['entities']
            if pred[0]['value'] in label:
                correct += 1
 
        self.write_result('Total', correct, len(vlsp))

    def test_time(self):
        tc_num = 1
        total_acc = 0
        total_len = 0
        for tc, ex_result in zip(self.time_tc, self.time_expected_result):
            acc = 0
            for test, label in zip(tc, ex_result):
                pred = self.extractor.extract_time(test)['entities']
                if pred[0]['value'] == label:
                    acc += 1
                # print(pred[0]['value'], label , acc,pred[0]['extractor'])
            self.write_result(tc_num, acc, len(tc))
            tc_num += 1
            total_len += len(tc)
            total_acc += (acc)
        for tc, ex_result in zip(self.time_spec_tc, self.time_spec_expected_result):
            acc = 0
            for test, label in zip(tc, ex_result):
                pred = self.extractor.extract_time(test)['entities']
                pred = pred[0]['value']
                now = datetime.datetime.now()
                hour = now.hour + int(label[0])
                minute = now.minute + int(label[1])
                if minute >= 60:
                    minute -= 60
                    hour += 1
                if hour - pred[0] == 0 and minute - pred[1] <= 1:
                    acc += 1
                print(pred, hour ,minute, acc)
            self.write_result(tc_num, acc, len(tc))
            tc_num += 1
            total_len += len(tc)
            total_acc += (acc)
        self.write_result('Total', total_acc, total_len)

    def test_date(self):
        correct = 0
        len_total_test = 0
        tc_num = 1
        for tc, ex_result in zip(self.date_tc, self.date_expected_result):
            len_total_test += len(tc)
            correct_tc = 0
            for test, target in zip(tc, ex_result):
                # print(test)
                pred = self.extractor.extract_date(test)
                # print(pred)
                entitites = pred['entities']
                result = []
                for i in entitites:
                    value = i['value']
                    result.extend(value)

                if result == target:
                    correct += 1
                    correct_tc += 1
            len_tc = len(tc)
            self.write_result(tc=tc_num, correct=correct_tc, total=len_tc)
            tc_num += 1

        acc = correct/len_total_test
        self.write_result(tc='Total', correct=correct, total=len_total_test)
        return acc

    def test_intent(self):
        pass

    def unit_test(self,features='pername'):
        assert features in ['pername','time','date']
        while True:
            utterance = str(input("Enter a sentence:"))
            if features == 'pername':
                result = self.extractor.extract_person_name(utterance)
            elif features == 'time':
                result = self.extractor.extract_time(utterance)
            elif features == 'date':
                result = self.extractor.extract_datetime(utterance)
            print(result)

    def test_template(self,tem_dir = './vi_nlp_core/testcase/template'):
        # Set up data
        template = read(tem_dir+'/template.txt')
        data_template = pd.read_csv(tem_dir + '/data_template.csv')
        person_data = [self.preprocessor.preprocess((str(x).lower())) for x in data_template['person'].tolist() if str(x) != 'nan']
        date_time_data = [self.preprocessor.preprocess(str(x)) for x in data_template['date/time'].tolist() if str(x)!='nan']
        date_data = [self.preprocessor.preprocess(str(x)) for x in data_template['date'].tolist() if str(x)!='nan']
        time_data = [self.preprocessor.preprocess(str(x)) for x in data_template['time'].tolist() if str(x) != 'nan']
            
        date_time_data.extend(date_data)
        date_time_data.extend(time_data)

        # list of generated data
        data = self.generate_data_from_template(template,data_template)
        # Testing 
        time_res = []
        pername_res = []
        date_res = []

        per_cnt = 0
        date_cnt = 0
        time_cnt = 0
        for i in tqdm(range(len(data))):
            pername = self.extractor.extract_person_name(data[i], mode='pattern', rt='relative')['entities'][0]['value']
            pername = self.preprocessor.preprocess(pername.lower())
            if pername in person_data:
                per_cnt += 1
            time = self.extractor.extract_time(data[i])['entities']
            if time != []:
                time = data[i][max(time[0]['start'] - self.error_factor,0):min(time[0]['end'] + self.error_factor,len(data[i]))]
                time = self.preprocessor.preprocess(str(time))
                for label in date_time_data:
                    if label in time or time in label:
                        time_cnt += 1
                        break
            try:
                date = self.extractor.extract_date(data[i])['entities']
                if date != []:
                    date = data[i][max(date[0]['start'] - self.error_factor,0):min(date[0]['end'] + self.error_factor,len(data[i]))]
                    date = self.preprocessor.preprocess(str(date))
                    for label in date_time_data:
                        if label in date or date in label:
                            date_cnt += 1
                            break
            except:
                date = []
                print(data[i])
            time_res.append(time)
            pername_res.append(pername)
            date_res.append(date)

        self.write_result('Person name Accuracy',per_cnt,len(data))
        self.write_result('Date Accuracy',date_cnt,len(data))
        self.write_result('Time Accuracy',time_cnt,len(data))
        df = pd.DataFrame({
            'utterance' : data,
            'pername' : pername_res,
            'time' : time_res,
            'date' : date_res,
        })
        df.to_csv('template_result.csv',index=False)
        
    def generate_data_from_template(self,template,data_template):

        data_pattern = {
            '[want]' : ['tôi muốn','muốn','tôi','cho tôi'],
            '[want_cancel]' : ['giúp','cho tôi','giúp cho tôi','giúp tôi'],
            '[pronoun]' : ['bs','bác sĩ','dr','bacsi'],
            '[date/time]' : [str(x) for x in data_template['date/time'].tolist() if str(x)!='nan'],
            '[date]' : [str(x) for x in data_template['date'].tolist() if str(x)!='nan'],
            '[time]' : [str(x) for x in data_template['time'].tolist() if str(x) != 'nan'],
            '[person]' :[str(x) for x in data_template['person'].tolist() if str(x) != 'nan'],
            '[khoa]' :  [str(x) for x in data_template['khoa'].tolist() if str(x) != 'nan']
        }

        global_verification = ['[want]','[pronoun]','[person]','[date/time]', '[khoa]', '[want_cancel]', '[time]','[date]']
        final_data = []
        final_intent = []

        final_pername = []
        final_date = []
        final_time = []
        for sentence in template:
            pattern = []
            generated_data = []
            expected_intent = []

            pername_label = []
            date_label = []
            time_label = []
            
            for token in sentence.split(' '):
                if token in global_verification:
                    pattern.append(data_pattern[token])
                else:
                    pattern.append([token])

            for r in itertools.product(*pattern):
                text = ' '.join(list(r))
                generated_data.append(text)
                if sentence.find('hủy') != -1:
                    expected_intent.append('cancel')
                else:
                    expected_intent.append('book')
            final_data.extend(generated_data)
            final_intent.extend(expected_intent)

        df = pd.DataFrame({
            'utterance' : final_data,
            'intent' : ['default'] * len(final_data),
            'expected_intent' : final_intent
        })
        df.to_csv('template_data.csv', index=False)

        return final_data

    def read_test_cases(self, path):
        pass

    def write_result(self, tc, correct, total):
        print(f"------------Test case {tc}------------")
        print(f"{colored('Pass','green')} {correct}")
        print(f"{colored('Fail','red')} {total-correct}")
        print(f"{colored('Total','green')} {correct}/{total}")



# source
# https://github.com/duyet/vietnamese-namedb
# https://github.com/thuy-le-ep/Vietnamese-data/tree/master/Vietnamese%20person%20names

import ahocorasick 
import itertools
import pickle
import json
import ast
import re

from vma_nlu.utils.util import save

def read(path):
    data = []
    with open(path, 'r') as f:
        for line in f:
            line = re.sub(' +', ' ',line)
            data.append(line.replace('\n', '').lower())
    return data

def build_full_name(
    fn_path = './first_name.txt',
    mn_path = './mid_name.txt',
    ln_path = './last_name.txt'
    ):
    # fn = read(fn_path)
    # mn = read(mn_path)
    # ln = read(ln_path)

    # mid + last name
    # ml = itertools.product(mn, ln)
    # ml = [' '.join(item) for item in ml]
    
    # first + mid + last name
    # fml =  itertools.product(fn, ml)
    # fml = [' '.join(item) for item in fml]
    # print("Saving first + mid + last name ...")
    # save(fml,'fullname2.txt')

    # first + first name
    # ff = itertools.product(fn, fn)
    # ff = [' '.join(item) for item in ff]

    # first + first + mid + last name
    ml = read('./midlastname.txt')
    ff = read('./firstfirstname.txt')
    ffml =  itertools.product(ff, ml)
    ffml = [' '.join(item) for item in ffml]

    # Save to txt 
    # ffml.extend(fml)
    # ffml = list(set(ffml))
    print("Saving first + first + mid + last name ...")
    save(ffml,'fullname2.txt')
def build_dict(
    path='./fullname.txt',
    path_out = './data/fullname.pkl'
    ):

    full_name = read(path)

    fullname_dict = ahocorasick.Automaton()

    idx = 0
    for key in full_name:
        fullname_dict.add_word(key, (idx, key))
        idx += 1

    fullname_dict.make_automaton()

    print('Saving dictionary ...')
    with open(path_out, 'wb') as file:
        pickle.dump(fullname_dict, file)
        

if __name__ == '__main__':
    # build_full_name()

    build_dict(
        path='./stopwords.txt',
        path_out = './data/stopwords.pkl'
    )


import os
from setuptools import setup, find_packages
from os import path

try:
    # pip >=20
    from pip._internal.network.session import PipSession
    from pip._internal.req import parse_requirements
except ImportError:
    try:
        # 10.0.0 <= pip <= 19.3.1
        from pip._internal.download import PipSession
        from pip._internal.req import parse_requirements
    except ImportError:
        # pip <= 9.0.3
        from pip.download import PipSession
        from pip.req import parse_requirements

long_description = ''

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session=False)

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# reqs is a list of requirement
# e.g. ['django==1.5.1', 'mezzanine==1.4.6']
try:
    reqs = [str(ir.req) for ir in install_reqs]
except:
    reqs = [str(ir.requirement) for ir in install_reqs]

VERSION = os.getenv('PACKAGE_VERSION', 'v0.0.1')[1:]

setup(
    name='vi_nlp_core',
    version='0.3.2',
    description='vi-core-nlp is a library that supports Vietnamese NER by pattern matching .',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    url='https://github.com/demdecuong/vi-core-nlp/tree/main',
    packages=find_packages(),
    author='minhnp et al.',
    author_email='nguyenphucminh2804@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    install_requires=reqs,
    keywords='vi_nlp_core',
    python_requires='>=3.6',
    py_modules=['vi_nlp_core'],

)


import re
import os

# from underthesea import ner
from vi_nlp_core.ner import date_matcher
from vi_nlp_core.utils.util import read_dict, get_ngram, tokenize
from vi_nlp_core.utils.pername_pattern import get_person_pattern, get_phone_pattern
from vi_nlp_core.utils.preproces import Preprocess
from vi_nlp_core.ner.date_matcher import DateMatcher
from vi_nlp_core.ner.time_matcher import TimeMatcher
from vi_nlp_core.ner.pername_matcher import PernameMatcher


ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
NAME_PATH = os.path.join(ROOT_DIR, 'data/fullname.pkl')
SW_PATH = os.path.join(ROOT_DIR, 'data/stopwords.pkl')
DATE_PATH = os.path.join(ROOT_DIR, 'data/dictionary_normalize_date.json')


class Extractor:
    def __init__(self, n_gram=4, load_dict=False):

        self.max_n_gram = n_gram

        self.phone_regex = get_phone_pattern()

        self.patternmatching_date = DateMatcher(DATE_PATH)

        self.time_extractor = TimeMatcher()

        self.pername_extractor = PernameMatcher(
            n_gram, NAME_PATH, SW_PATH, load_dict)

        self.preprocessor = Preprocess()

        # self.pername_extractor_deeplearning = Inference()
    def extract_ner(self, utterance):
        '''
        Extract NER based on given intent
        Input:
            utterance   -   string
            return_value    return value only 
        Return
        '''
        result = []
        person_name = self.extract_person_name(utterance)['entities']

        if person_name != []:
            result.extend(person_name)
        time = self.extract_time(utterance)['entities']
        if time != []:
            result.extend(time)
        date = self.extract_date(utterance)['entities']
        if date != []:
            result.extend(date)
        return result

    def extract_ner_old(self, utterance, intent):
        '''
        Extract NER based on given intent
        Input:
            utterance   -   string
            intent      -   string 
        Return
        '''
        #assert intent in ['greet','goodbye','thank','book_appt','change_appt','cancel_appt','inform','agree','disagree']
        try:
            if intent in ['inform', 'book_appt', 'change_appt', 'cancel_appt']:
                result = []
                person_name = self.extract_person_name(utterance)['entities']
                print(person_name)
                if person_name != []:
                    result.extend(person_name)
                time = self.extract_time(utterance)['entities']
                print(time)
                if time != []:
                    result.extend(time)
                date = self.extract_date(utterance)['entities']
                print(date)
                if date != []:
                    result.extend(date)
                return result
                # return {
                #         'entities' : result
                #     }
            else:
                return []
        except Exception as ex:
            print(ex)
            return []

    def extract_time(self, utterance, return_value = False):
        '''
        Input:
            utterance :  string
            return_value : return value only
        Output:
            return value of (hour,minute)
        '''
        result = self.time_extractor.extract_time(utterance)
        if return_value:
            final_result = []
            for ent in result['entities']:
                final_result.append(ent['value'])
            return final_result
        return result

    def extract_date(self, utterance, return_value = False):
        '''
        Input:
            utterance : input string
            return_value : return value only
        Output:
            return tuple of (dow,dd,mm,yyyy)
        '''
        result = self.patternmatching_date.extract_date(utterance)
        if return_value:
            final_result = []
            for ent in result['entities']:
                final_result.append(ent['value'])
            return final_result
        return result

    def extract_person_name(self, utterance, return_value= False):
        '''
        Input:
            input string
        Output:
            return a json with 'value' is a string 
        '''
        result = self.pername_extractor.extract_person_name(utterance)
        if return_value:
            final_result = []
            for ent in result['entities']:
                final_result.append(ent['value'])
            return final_result
        return result

    def extract_phone_num(self, utterance):
        '''
        Use regex for pattern matching (nha mang + remain numbers)
        Input:
            utterance  - string
        Return:
            string      :   phone number
            'Invalid'   :   otherwise
        '''
        result = re.findall(self.phone_regex, utterance)
        if result == []:
            return 'Invalid'
        else:
            nha_mang = result[0][0]
            remain = result[0][1]
            return nha_mang + remain
    def extract_all(self, utterance):
        time = self.extract_time(utterance)
        date = self.extract_date(utterance)

        person_name = self.extract_person_name(utterance)
        return time, date, person_name  